# mil/bag.py
from __future__ import annotations
from dataclasses import dataclass
from typing import List, Sequence, Optional
import numpy as np
import numpy.typing as npt

Label = float  # or int; we'll store 0/1 or -1/+1; convert as needed downstream

@dataclass
class Bag:
    """A bag of instances with a bag-level label and per-instance (0/1) flags."""
    X: npt.NDArray[np.float64]            # shape (n_i, d)
    y: Label                              # bag label (e.g., 0/1 or -1/+1)
    intra_bag_label: Optional[npt.NDArray[np.float64]] = None  # shape (n_i,), 0/1

    def __post_init__(self):
        self.X = np.asarray(self.X, dtype=float)
        if self.X.ndim != 2:
            raise ValueError("Bag.X must be 2D (n_i, d).")
        n_i = self.X.shape[0]
        if self.intra_bag_label is None:
            # default: all ones
            self.intra_bag_label = np.ones(n_i, dtype=float)
        else:
            self.intra_bag_label = np.asarray(self.intra_bag_label, dtype=float).ravel()
            if self.intra_bag_label.shape[0] != n_i:
                raise ValueError("intra_bag_label length must match number of instances.")

    @property
    def n(self) -> int: return self.X.shape[0]

    @property
    def d(self) -> int: return self.X.shape[1]

    @property
    def mask(self) -> npt.NDArray[np.float64]:
        """Safe 0/1 mask (clips negatives and >1)."""
        return np.clip(self.intra_bag_label, 0.0, 1.0)

    def positives(self) -> npt.NDArray[np.int64]:
        """Indices of instances with intra_bag_label == 1."""
        return np.flatnonzero(self.mask >= 0.5)

    def negatives(self) -> npt.NDArray[np.int64]:
        """Indices of instances with intra_bag_label == 0."""
        return np.flatnonzero(self.mask < 0.5)

@dataclass
class BagDataset:
    bags: List[Bag]

    @staticmethod
    def from_arrays(
        bags: Sequence[np.ndarray],
        y: Sequence[float],
        intra_bag_labels: Sequence[np.ndarray] | None = None
    ) -> "BagDataset":
        if intra_bag_labels is None:
            intra_bag_labels = [None] * len(bags)
        if len(bags) != len(y) or len(bags) != len(intra_bag_labels):
            raise ValueError("bags, y, intra_bag_labels must have same length.")
        return BagDataset([
            Bag(X=b, y=float(lbl), intra_bag_label=ibl)
            for b, lbl, ibl in zip(bags, y, intra_bag_labels)
        ])

    def split_by_label(self) -> tuple[list[Bag], list[Bag]]:
        return self.positive_bags(), self.negative_bags()

    def Xy(self) -> tuple[list[np.ndarray], np.ndarray, list[np.ndarray]]:
        Xs = [b.X for b in self.bags]
        ys = np.asarray([b.y for b in self.bags], dtype=float)
        masks = [b.mask for b in self.bags]
        return Xs, ys, masks

    def positive_bags(self) -> list[Bag]:
        return [b for b in self.bags if float(b.y) > 0.0]

    def negative_bags(self) -> list[Bag]:
        return [b for b in self.bags if float(b.y) <= 0.0]

    def positive_instances(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Instances from positive bags with intra_bag_label == 1.
        Returns:
            X_pos: (N, d)
            bag_index: (N,) indices into self.bags (original positions)
        """
        Xs, bag_idx = [], []
        for i, b in enumerate(self.bags):          # iterate original list!
            if float(b.y) <= 0.0:
                continue
            mask = b.mask >= 0.5
            if np.any(mask):
                Xs.append(b.X[mask])
                bag_idx.extend([i] * int(mask.sum()))
        if not Xs:
            d = self.bags[0].d if self.bags else 0
            return np.zeros((0, d)), np.array([], dtype=int)
        return np.vstack(Xs), np.array(bag_idx, dtype=int)

    def negative_instances(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Instances from:
        - all negative bags (all instances),
        - plus from positive bags where intra_bag_label == 0.
        Returns:
            X_neg: (M, d)
            bag_index: (M,) indices into self.bags (original positions)
        """
        Xs, bag_idx = [], []
        for i, b in enumerate(self.bags):          # iterate original list!
            if float(b.y) <= 0.0:
                # take all instances
                if b.X.shape[0] > 0:
                    Xs.append(b.X)
                    bag_idx.extend([i] * b.X.shape[0])
            else:
                # positive bag: only intra == 0
                mask = b.mask < 0.5
                if np.any(mask):
                    Xs.append(b.X[mask])
                    bag_idx.extend([i] * int(mask.sum()))
        if not Xs:
            d = self.bags[0].d if self.bags else 0
            return np.zeros((0, d)), np.array([], dtype=int)
        return np.vstack(Xs), np.array(bag_idx, dtype=int)

    def negative_bags_as_singletons(self) -> list[Bag]:
        '''
        Transforms all negative bags into singleton bags, by flattening each bag  (b, n, d) -> (b x n, d)
        '''
        singletons: list[Bag] = []
        for b in self.negative_bags():
            for j in range(b.n):
                singletons.append(Bag(X=b.X[j:j+1, :], y=-1.0))
        return singletons

    def positive_bags_as_singletons(self) -> list[Bag]:
        '''
        Transforms all positive bags into singleton bags, by flattening each bag  (b, n, d) -> (b x n, d)
        '''
        singletons: list[Bag] = []
        for b in self.positive_bags():
            for j in range(b.n):
                singletons.append(Bag(X=b.X[j:j+1, :], y=1.0))
        return singletons

    @property
    def num_pos_instances(self) -> int:
        return sum(b.n for b in self.positive_bags())

    @property
    def num_neg_instances(self) -> int:
        return sum(b.n for b in self.negative_bags())

    @property
    def num_instances(self) -> int:
        return self.num_pos_instances + self.num_neg_instances

    @property
    def num_bags(self) -> int:
        return len(self.bags)
    
    @property
    def num_pos_bags(self) -> int:
        return len(self.positive_bags())

    @property
    def num_neg_bags(self) -> int:
        return len(self.negative_bags())


# sparse_mil/kernels.py
# Kernel Implementation for the single-instance cases
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Union
import numpy as np
import numpy.typing as npt
import logging

log = logging.getLogger("sparse_mil.kernels")


# ---------------- Utilities 
def _sqeuclidean(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    Xn = (X * X).sum(1)[:, None]
    Yn = (Y * Y).sum(1)[None, :]
    return Xn + Yn - 2.0 * (X @ Y.T)

# ---------------- Base class 
class BaseKernel(ABC):
    """Minimal kernel interface for the single-instance kernels: fit (optional) + __call__."""
    def fit(self, X: npt.NDArray[np.float64]) -> "BaseKernel":
        return self
    @abstractmethod
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        ...

# ----------------  Single-instance kernels 
@dataclass
class Linear(BaseKernel):
    def fit(self, X: npt.NDArray[np.float64]) -> "Linear":
        log.warning("Linear kernel has no parameters to fit.")
        return self
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> np.ndarray:
        return X @ Y.T

@dataclass
class RBF(BaseKernel):
    gamma: Optional[float] = None  # if None, set to 1/d in fit()
    def fit(self, X: npt.NDArray[np.float64]) -> "RBF":
        if self.gamma is None:
            self.gamma = 1.0 / X.shape[1]
        return self
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> np.ndarray:
        assert self.gamma is not None, "Call .fit(X) or set gamma."
        return np.exp(-self.gamma * _sqeuclidean(X, Y))

@dataclass
class Polynomial(BaseKernel):
    degree: int = 3
    gamma: Optional[float] = None
    coef0: float = 0.0
    def fit(self, X: npt.NDArray[np.float64]) -> "Polynomial":
        if self.gamma is None:
            self.gamma = 1.0 / X.shape[1]
        return self
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> np.ndarray:
        assert self.gamma is not None
        return (self.gamma * (X @ Y.T) + self.coef0) ** self.degree

@dataclass
class Sigmoid(BaseKernel):
    gamma: Optional[float] = None
    coef0: float = 0.0
    def fit(self, X: npt.NDArray[np.float64]) -> "Sigmoid":
        if self.gamma is None:
            self.gamma = 1.0 / X.shape[1]
        return self
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> np.ndarray:
        assert self.gamma is not None
        return np.tanh(self.gamma * (X @ Y.T) + self.coef0)

@dataclass
class Precomputed(BaseKernel):
    """Use when a Gram matrix is already built; ignores X,Y and returns K (shape checked by caller)."""
    K: np.ndarray
    def fit(self, X: npt.NDArray[np.float64]) -> "Precomputed":
        log.warning("Precomputed kernel has no parameters to fit.")
        return self
    def __call__(self, X: npt.NDArray[np.float64], Y: npt.NDArray[np.float64]) -> np.ndarray:
        return self.K

# ---------------- Combinators 
@dataclass
class Scale(BaseKernel):
    a: float
    k: BaseKernel
    def fit(self, X): self.k.fit(X); return self
    def __call__(self, X, Y): return self.a * self.k(X, Y)

@dataclass
class Sum(BaseKernel):
    k1: BaseKernel
    k2: BaseKernel
    def fit(self, X): self.k1.fit(X); self.k2.fit(X); return self
    def __call__(self, X, Y): return self.k1(X, Y) + self.k2(X, Y)

@dataclass
class Product(BaseKernel):
    k1: BaseKernel
    k2: BaseKernel
    def fit(self, X): self.k1.fit(X); self.k2.fit(X); return self
    def __call__(self, X, Y): return self.k1(X, Y) * self.k2(X, Y)

@dataclass
class Normalize(BaseKernel):
    """Cosine-style normalization: Kxy / sqrt(Kxx * Kyy)."""
    k: BaseKernel
    eps: float = 1e-12
    def fit(self, X): self.k.fit(X); return self
    def __call__(self, X, Y):
        Kxy = self.k(X, Y)
        Kxx = np.sqrt(np.maximum(np.diag(self.k(X, X)), self.eps))[:, None]
        Kyy = np.sqrt(np.maximum(np.diag(self.k(Y, Y)), self.eps))[None, :]
        return Kxy / (Kxx * Kyy + self.eps)

# ---------------- Registry & Resolver 
# Used for the type checking
KernelType = Union[
    str,                                     # "linear", "rbf", "poly", "sigmoid", "precomputed"
    BaseKernel,                               # already-constructed kernel
    Callable[[np.ndarray, np.ndarray], np.ndarray],  # raw callable k(X, Y)
]

def get_kernel(spec: KernelType, **kwargs) -> BaseKernel:
    """
    Normalize various 'kernel=' inputs into a BaseKernel object.
    kwargs are used only when spec is a string (e.g., gamma, degree, coef0, K).
    """
    # already an object?
    if isinstance(spec, BaseKernel):
        return spec

    # plain callable? wrap so it behaves like a BaseKernel
    if callable(spec) and not isinstance(spec, str):
        class _CallableKernel(BaseKernel):
            def __call__(self, X, Y): return np.asarray(spec(X, Y), dtype=float)
        return _CallableKernel()

    # by name
    name = str(spec).lower()
    if name == "precomputed":
        if "K" not in kwargs:
            raise ValueError("get_kernel('precomputed', K=...) requires the Gram matrix K.")
        return Precomputed(np.asarray(kwargs["K"], dtype=float))

    registry: Dict[str, BaseKernel] = {
        "linear":  Linear(),
        "rbf":     RBF(kwargs.get("gamma")),
        "poly":    Polynomial(kwargs.get("degree", 3), kwargs.get("gamma"), kwargs.get("coef0", 0.0)),
        "sigmoid": Sigmoid(kwargs.get("gamma"), kwargs.get("coef0", 0.0)),
    }
    if name in registry:
        return registry[name]

    raise ValueError(f"Unknown kernel spec: {spec!r}")

__all__ = [
    "BaseKernel",
    "Linear", "RBF", "Polynomial", "Sigmoid", "Precomputed",
    "Scale", "Sum", "Product", "Normalize",
    "KernelType", "get_kernel",
]


# sparse_mil/smil.py
from __future__ import annotations
from typing import Optional, Sequence, List
import numpy as np
import numpy.typing as npt

from .nsk import NSK
from .bag import Bag, BagDataset
from .bag_kernels import WeightedMeanBagKernel
from .kernels import Linear
from .quadprog import quadprog_gurobi


class sMIL(NSK):
    """
    Sparse MIL (Bunescu & Mooney, 2007) implemented on top of NSK.

    Training set:
      - Every negative *instance* becomes its own 1-instance bag (label -1).
      - Every positive bag stays a bag (label +1).

    Dual tweaks:
      - Linear term for positive bags: f_j = 2/|B_j| - 1
      - Box constraints: iC for negatives, bC for positives (scaled if scale_C)

    Notes:
      - By default we ignore intra-bag labels (uniform instance weights).
      - Use your NSK's bag kernel; mean aggregator with normalizer="none" is a good default.
    """

    @staticmethod
    def _coerce_bags_and_labels(
        bags: Sequence[Bag] | BagDataset | Sequence[np.ndarray],
        y: Optional[npt.NDArray[np.float64]] = None
    ) -> tuple[List[Bag], npt.NDArray[np.float64]]:
        # Reuse NSK’s accepted inputs; prefer Bag.y when available
        if isinstance(bags, BagDataset):
            bl = list(bags.bags)
            ya = np.asarray([b.y for b in bl], dtype=float)
            return bl, ya
        if len(bags) > 0 and isinstance(bags[0], Bag):  # type: ignore[index]
            bl = list(bags)  # type: ignore[assignment]
            ya = np.asarray([b.y for b in bl], dtype=float) if y is None else np.asarray(y, dtype=float).ravel()
            if y is not None and ya.shape[0] != len(bl):
                raise ValueError("Length of y must equal number of bags.")
            return bl, ya
        if y is None:
            raise ValueError("When passing raw arrays for bags, you must also pass y.")
        bl = [Bag(X=np.asarray(b, dtype=float), y=float(lbl)) for b, lbl in zip(bags, y)]
        return bl, np.asarray(y, dtype=float).ravel()

    @staticmethod
    def _build_init_training(bags: List[Bag]) -> tuple[List[Bag], npt.NDArray[np.float64], int, int]:
        # negative instances -> singleton bags
        negative_instances: List[Bag] = [
            Bag(X=b.X[j:j+1, :], y=-1.0)
            for b in bags if float(b.y) <= 0.0
            for j in range(b.n)
        ]
        # positive bags unchanged
        positive_bags: List[Bag] = [b for b in bags if float(b.y) > 0.0]

        train_bags = negative_instances + positive_bags
        S_n = len(negative_instances)      # # negative instances
        B_p = len(positive_bags)           # # positive bags
        y_train = np.hstack([-np.ones(S_n), np.ones(B_p)])
        return train_bags, y_train, S_n, B_p

    def fit(self, bags: Sequence[Bag] | BagDataset | Sequence[np.ndarray],
            y: Optional[npt.NDArray[np.float64]] = None) -> "sMIL":
        # 1) coerce inputs and build sMIL training set
        init_bags, _ = self._coerce_bags_and_labels(bags, y)
        train_bags, y_train, S_n, B_p = self._build_init_training(init_bags)
        if not train_bags:
            raise ValueError("No training data after sMIL transformation.")
        self.bags_ = train_bags

        # map to {-1,+1} and store classes
        classes = np.unique(y_train)
        if classes.size != 2:
            raise ValueError("Binary classification only.")
        self.classes_ = classes.astype(float)
        Y = np.where(y_train == classes[0], -1.0, 1.0)
        self.y_ = Y

        # 2) bag kernel Gram
        bk = self._ensure_bag_kernel()
        bk.fit(train_bags)
        K = bk(train_bags, train_bags)                     # (n, n)

        # 3) QP pieces
        H = (Y[:, None] * Y[None, :]) * K
        H = 0.5 * (H + H.T)
        n = len(train_bags)
        f = -np.ones(n, dtype=float)
        sizes = np.array([b.n for b in train_bags[S_n:]], dtype=float)
        f[S_n:] = (2.0 / np.maximum(sizes, 1.0)) - 1.0

        Aeq = Y.reshape(1, -1)
        beq = np.array([0.0], dtype=float)

        # per-variable box constraints
        if self.scale_C:
            iC = float(self.C) / max(S_n, 1)
            bC = float(self.C) / max(B_p, 1)
        else:
            iC = float(self.C)
            bC = float(self.C)
        lb = np.zeros(n, dtype=float)
        ub = np.concatenate([np.full(S_n, iC), np.full(B_p, bC)]).astype(float)

        # 4) solve
        alpha, _ = quadprog_gurobi(H, f, Aeq, beq, lb, ub, verbose=self.verbose)
        self.alpha_ = alpha

        # 5) SVs + intercept (dual)
        sv_mask = alpha > self.tol
        self.support_ = np.flatnonzero(sv_mask).astype(int)
        self.support_vectors_ = [train_bags[i] for i in self.support_]
        self.dual_coef_ = (alpha[sv_mask] * Y[sv_mask]).reshape(1, -1)

        caps = np.concatenate([np.full(S_n, iC), np.full(B_p, bC)]).astype(float)
        on_margin = (alpha > self.tol) & (alpha < (caps - self.tol))
        if not np.any(on_margin):
            on_margin = sv_mask
        b_vals = Y[on_margin] - (alpha * Y) @ K[:, on_margin]
        self.intercept_ = float(np.mean(b_vals)) if b_vals.size else 0.0

        # 6) optional primal recovery (linearizable case: Linear + WeightedMean + p==1)
        if isinstance(bk, WeightedMeanBagKernel) and isinstance(bk.inst_kernel, Linear) and abs(bk.p - 1.0) < 1e-12:
            # φ(B) consistent with your NSK (uniform weights, chosen normalizer)
            Z = np.stack([self._phi(b, normalizer=bk.normalizer) for b in train_bags], axis=0)  # (n, d)
            self.coef_ = (self.alpha_ * self.y_) @ Z
            use = on_margin if np.any(on_margin) else sv_mask
            if np.any(use):
                self.intercept_ = float(np.mean(self.y_[use] - Z[use] @ self.coef_))
        else:
            self.coef_ = None  # keep dual intercept

        return self



from __future__ import annotations
import numpy as np
from numpy import typing as npt
from typing import Tuple, Optional
from dataclasses import dataclass
import logging

log = logging.getLogger("quadprog")

@dataclass
class Objective:
    """Container for quadratic/linear objective parts."""

    quadratic: float
    linear: float

    @property
    def objective(self) -> float:
        return self.quadratic + self.linear

def quadprog_gurobi(
    H: npt.NDArray[np.float64],     
    f: npt.NDArray[np.float64],       
    Aeq: Optional[npt.NDArray[np.float64]], 
    beq: Optional[npt.NDArray[np.float64]],  
    lb: npt.NDArray[np.float64],     
    ub: npt.NDArray[np.float64],      
    verbose: bool = False,           
)-> Tuple[npt.NDArray[np.float64], "Objective"]:
    """
    Solve the quadratic program:

        minimize   0.5 * αᵀ H α + fᵀ α
        subject to Aeq α = beq
                   lb ≤ α ≤ ub

    Args:
        H: (n, n) quadratic term matrix in 0.5 * αᵀ H α
        f: (n,) linear term vector in fᵀ α , usually f = -1
        Aeq: (m, n) equality constraint matrix, usually yᵀ
        beq: (m,) equality constraint rhs, usually 0
        lb: (n,) lower bound vector, usually 0
        ub: (n,) upper bound vector, usually C
        verbose: If True, print solver logs

    Returns:
        α*: Optimal solution vector
        Objective: quadratic and linear parts of the optimum
    """

    # ---- VALIDATION
    try:  # pragma: no cover - gurobi may be unavailable in some envs
        import gurobipy as gp
    except Exception as exc:  # pragma: no cover
        raise ImportError("gurobipy is required for solver='gurobi'") from exc
    if (Aeq is None) ^ (beq is None):
        raise ValueError("Aeq and beq must both be None or both be provided.")

    n = H.shape[0]
    # ---- COERCION and VALIDATION
    H = np.asarray(H, dtype=float)
    f = np.asarray(f, dtype=float)
    lb = np.asarray(lb, dtype=float)
    ub = np.asarray(ub, dtype=float)
    if H.shape != (n, n):
        raise ValueError(f"H must be (n,n); got {H.shape}")
    if f.shape != (n,):
        raise ValueError(f"f must be (n,); got {f.shape}")
    if lb.shape != (n,) or ub.shape != (n,):
        raise ValueError("lb and ub must be length-n vectors.")
    if Aeq is not None:
        Aeq = np.asarray(Aeq, dtype=float)
        beq = np.asarray(beq, dtype=float)
        if Aeq.shape[1] != n:
            raise ValueError(f"Aeq must have n columns; got {Aeq.shape}")
        if beq.shape != (Aeq.shape[0],):
            raise ValueError(f"beq must match Aeq rows; got beq {beq.shape}, Aeq {Aeq.shape}")

    # ---- STABILITY
    H = 0.5 * (H + H.T)

    # --- Optimization
    model = gp.Model()
    if not verbose:
        model.Params.OutputFlag = 0
        
    # should we include? 
    # model.Params.NumericFocus = 1
    # model.Params.OptimalityTol = 1e-8
    # model.Params.BarConvTol = 1e-10

    x = model.addMVar(n, lb=lb, ub=ub, name="alpha")
    obj = 0.5 * (x @ H @ x) + f @ x
    model.setObjective(obj, gp.GRB.MINIMIZE)

    if Aeq is not None:
        model.addConstr(Aeq @ x == beq, name="eq")
    model.optimize()

    if model.Status != gp.GRB.OPTIMAL:  # pragma: no cover - defensive
        log.warning(RuntimeError(
            f"Gurobi optimization failed with status {model.Status}"))


    xstar = np.asarray(x.X, dtype=float)
    quadratic = float(0.5 * xstar.T @ H @ xstar)
    linear = float(f.T @ xstar)
    return xstar, Objective(quadratic, linear)


# sparse_mil/bag_kernels.py
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Literal
import numpy as np
import numpy.typing as npt

from .bag import Bag
from .kernels import BaseKernel, Linear

# ---------- Normalizer Utils 
_NormalizerName = Literal["none", "average", "featurespace"]

# ---------- Base Multiple-instance Kernel
class BaseBagKernel(ABC):
    '''Base class for the Multiple-instance (bags) kernel'''
    def fit(self, bags: List[Bag]) -> "BaseBagKernel":
        return self
    @abstractmethod
    def __call__(self, bags_X: List[Bag], bags_Y: List[Bag]) -> npt.NDArray[np.float64]:
        ...

def _effective_count(b: Bag) -> float:
    '''Counts the effective instances in a bag (relevant only if classifier or kernel uses the intra_bag_labels)'''
    s = float(b.mask.sum())
    return s if s > 0.0 else max(1.0, float(b.n))
    
@dataclass
class WeightedMeanBagKernel(BaseBagKernel):
    """
    K(Bi,Bj) = [ (w_i^T k(Bi,Bj) w_j) ] ** p / ( norm(Bi) * norm(Bj) ),
      where w_i, w_j are instance weights normalized to sum=1
      (fallback to uniform if a bag’s mask sums to 0).

    normalizer: (default is 'average')
      - "none"       -> norm(B)=1
      - "average"    -> norm(B)=sum(mask) (fallback to bag size)
      - "featurespace" -> norm(B)=sqrt(w^T k(X,X) w)
                          (fast for Linear via ||weighted_mean||)

    By default (use_intra_labels=False):
      - w_i are UNIFORM weights (1/n_i), i.e., intra labels are IGNORED.
      - normalizer defaults to "none" (no extra scaling).
    """
    inst_kernel: BaseKernel
    normalizer: _NormalizerName = "average"
    p: float = 1.0
    use_intra_labels: bool = False
    fast_linear: bool = True

    def fit(self, bags: List[Bag]) -> "WeightedMeanBagKernel":
        # If the instance kernel needs defaults (e.g. gamma), fit it on a few instances.
        # We can use the first bag to infer dimensionality.
        for b in bags:
            if b.n > 0:
                self.inst_kernel.fit(b.X)
                break
        return self

    # ---- helpers ----
    def _weights(self, b: Bag) -> npt.NDArray[np.float64]:
        if b.n == 0:
            return np.zeros((0,), dtype=float)

        if not self.use_intra_labels:
            # 👇 KEY CHANGE:
            # If "average" normalizer is requested, use UNNORMALIZED weights (all ones)
            # so the numerator is a SUM; the denominator will divide by counts -> net MEAN.
            if self.normalizer == "average":
                return np.ones(b.n, dtype=float)
            # otherwise, keep the plain mean
            return np.full(b.n, 1.0 / b.n, dtype=float)

        # use intra-bag mask (normalized to sum=1; fallback to uniform)
        w = b.mask.astype(float)
        s = float(w.sum())
        if s <= 0.0:
            return np.full(b.n, 1.0 / b.n, dtype=float)
        return w / s

    def _norms(self, bags: List[Bag]) -> npt.NDArray[np.float64]:
        if self.normalizer == "none":
            return np.ones(len(bags), dtype=float)

        if self.normalizer == "average":
            # 👇 CONSISTENT with _weights above:
            # - no intra labels: denominator = bag size
            # - with intra labels: denominator = effective positive count
            if not self.use_intra_labels:
                return np.array([max(b.n, 1) for b in bags], dtype=float)
            else:
                return np.array([max(float(b.mask.sum()), 1.0) for b in bags], dtype=float)

        # featurespace
        if self.fast_linear and isinstance(self.inst_kernel, Linear):
            means = np.stack(
                [self._weighted_mean(b) if b.n else np.zeros((bags[0].d,), dtype=float)
                 for b in bags],
                axis=0,
            )
            return np.linalg.norm(means, axis=1).clip(min=1e-12)
        norms = np.empty(len(bags), dtype=float)
        for i, b in enumerate(bags):
            if b.n == 0:
                norms[i] = 1.0
                continue
            w = self._weights(b)
            G = self.inst_kernel(b.X, b.X)
            val = float(w @ G @ w)
            norms[i] = np.sqrt(max(val, 1e-12))
        return norms

    def _pair_value(self, Bi: Bag, Bj: Bag) -> float:
        if Bi.n == 0 or Bj.n == 0:
            return 0.0
        wi = self._weights(Bi)
        wj = self._weights(Bj)
        if self.fast_linear and isinstance(self.inst_kernel, Linear):
            mi = (wi[None, :] @ Bi.X).ravel()
            mj = (wj[None, :] @ Bj.X).ravel()
            val = float(mi @ mj)
        else:
            G = self.inst_kernel(Bi.X, Bj.X)
            val = float(wi @ G @ wj)
        if self.p != 1.0:
            val = float(np.power(max(val, 0.0), self.p))
        return val

    def __call__(self, bags_X: List[Bag], bags_Y: List[Bag]) -> npt.NDArray[np.float64]:
        nX, nY = len(bags_X), len(bags_Y)
        K = np.empty((nX, nY), dtype=float)
        same = bags_X is bags_Y
        norms_X = self._norms(bags_X)
        norms_Y = norms_X if same else self._norms(bags_Y)
        for i, Bi in enumerate(bags_X):
            j0 = i if same else 0
            for j in range(j0, nY):
                val = self._pair_value(Bi, bags_Y[j])
                denom = norms_X[i] * norms_Y[j]
                kij = val / denom if denom > 0.0 else 0.0
                K[i, j] = kij
                if same and j != i:
                    K[j, i] = kij
        return K
    
# ---------- Precomputed bag kernel ----------
@dataclass
class PrecomputedBagKernel(BaseBagKernel):
    K: npt.NDArray[np.float64]
    def __call__(self, bags_X: List[Bag], bags_Y: List[Bag]) -> npt.NDArray[np.float64]:
        # Caller must pass consistent ordering to match K
        return np.asarray(self.K, dtype=float)

# ---------- Simple factory ----------
def make_bag_kernel(
    inst_kernel: BaseKernel,
    *,
    normalizer: _NormalizerName = "none",
    p: float = 1.0,
    use_intra_labels: bool = False,
    fast_linear: bool = True,
) -> WeightedMeanBagKernel:
    return WeightedMeanBagKernel(
        inst_kernel=inst_kernel,
        normalizer=normalizer,
        p=p,
        use_intra_labels=use_intra_labels,
        fast_linear=fast_linear,
    )


__all__ = [
    "WeightedMeanBagKernel",
    "PrecomputedBagKernel",
    "make_bag_kernel",
    "BaseBagKernel"
]


# sparse_mil/nsk.py
from __future__ import annotations
from typing import Optional, Sequence, List, Literal
import numpy as np
import numpy.typing as npt

from .svm import SVM                   
from .bag import Bag, BagDataset
from .kernels import get_kernel, KernelType, Linear
from .bag_kernels import BaseBagKernel, make_bag_kernel, WeightedMeanBagKernel
from .quadprog import quadprog_gurobi

class NSK(SVM):
    """
    Normalized Set Kernel SVM on bags of instances.
    Builds a bag-level Gram matrix with a bag kernel, then solves the standard SVM dual.
    """

    def __init__(
        self,
        C: float = 1.0,
        # If bag_kernel is None, we'll build one from this instance-kernel spec:
        base_kernel: KernelType = "linear",
        gamma: Optional[float] = None,
        degree: int = 3,
        coef0: float = 0.0,
        *,
        # Bag kernel settings:
        # only used if one does not specify bag_kernel
        normalizer: Literal["none", "average", "featurespace"] = "none",
        p: float = 1.0,
        use_intra_labels: bool = False,
        fast_linear: bool = True,
        bag_kernel: Optional[BaseBagKernel] = None,
        # Solver / SVM settings:
        scale_C: bool = True,
        tol: float = 1e-6,
        verbose: bool = False,
    ):
        # parent SVM stores common attrs; kernel arg unused here
        super().__init__(C=C, kernel="linear", gamma=gamma, degree=degree,
                         coef0=coef0, tol=tol, verbose=verbose)
        self.scale_C = scale_C

        # How to build the bag kernel (if not provided)
        self.base_kernel = base_kernel
        self.normalizer = normalizer
        self.p = p
        self.use_intra_labels = use_intra_labels
        self.fast_linear = fast_linear
        self.bag_kernel: Optional[BaseBagKernel] = bag_kernel

        # Fitted state
        self.bags_: Optional[List[Bag]] = None  # training bags (ordering matters)

    # ---------- helpers ----------
    @staticmethod
    def _coerce_bags_and_labels(
        bags: Sequence[Bag] | BagDataset | Sequence[np.ndarray],
        y: Optional[npt.NDArray[np.float64]] = None
    ) -> tuple[List[Bag], npt.NDArray[np.float64]]:
        # BagDataset
        if isinstance(bags, BagDataset):
            blist = list(bags.bags)
            y_arr = np.asarray([b.y for b in blist], dtype=float)
            return blist, y_arr
        # List[Bag]
        if len(bags) > 0 and isinstance(bags[0], Bag):  # type: ignore[index]
            blist = list(bags)  # type: ignore[assignment]
            y_arr = np.asarray([b.y for b in blist], dtype=float) if y is None else np.asarray(y, dtype=float).ravel()
            if y is not None and y_arr.shape[0] != len(blist):
                raise ValueError("Length of y must equal number of bags.")
            return blist, y_arr
        # Raw arrays + y (bags without intra labels -> all uniform weights)
        if y is None:
            raise ValueError("When passing raw arrays for bags, you must also pass y.")
        blist = [Bag(X=np.asarray(b, dtype=float), y=float(lbl)) for b, lbl in zip(bags, y)]
        return blist, np.asarray(y, dtype=float).ravel()

    def _ensure_bag_kernel(self) -> BaseBagKernel:
        if self.bag_kernel is not None:
            return self.bag_kernel
        # Build instance kernel, then lift to a bag kernel
        inst_k = get_kernel(self.base_kernel, gamma=self.gamma, degree=self.degree, coef0=self.coef0)
        # Note: bag_kernels make normalized weights (means) in the numerator,
        # so to avoid double normalization we default to normalizer="none".
        self.bag_kernel = make_bag_kernel(
            inst_kernel=inst_k,
            normalizer=self.normalizer,
            p=self.p,
            use_intra_labels=self.use_intra_labels,
            fast_linear=self.fast_linear,
        )
        # Fit once to allow instance kernel to set defaults (e.g., gamma)
        # (bag_k.fit looks at the first non-empty bag for dimensionality)
        return self.bag_kernel
    
    def _can_linearize(self, bk) -> bool:
        # recover w only for linear instance kernel + WeightedMeanBagKernel + p=1
        return (
            isinstance(bk, WeightedMeanBagKernel)
            and isinstance(bk.inst_kernel, Linear)
            and abs(bk.p - 1.0) < 1e-12
        )

    # ---------- sklearn-style API ----------
    def fit(self, bags: Sequence[Bag] | BagDataset | Sequence[np.ndarray],
            y: Optional[npt.NDArray[np.float64]] = None) -> "NSK":
        bag_list, y_arr = self._coerce_bags_and_labels(bags, y)
        if len(bag_list) == 0:
            raise ValueError("No bags provided.")
        self.bags_ = bag_list

        # Map labels to {-1, +1} (store original classes)
        classes = np.unique(y_arr)
        if classes.size != 2:
            raise ValueError("Binary classification only—y must have exactly two classes.")
        self.classes_ = classes.astype(float)
        Y = np.where(y_arr == classes[0], -1.0, 1.0)
        self.y_ = Y

        # Bag kernel -> Gram
        bk = self._ensure_bag_kernel()
        bk.fit(bag_list)
        K = bk(bag_list, bag_list)  # (n_bags, n_bags)

        # Build dual QP (same as SVM)
        H = (Y[:, None] * Y[None, :]) * K
        H = 0.5 * (H + H.T)                   # symmetrize for safety
        n = len(bag_list)
        f = -np.ones(n, dtype=float)
        Aeq = Y.reshape(1, -1)
        beq = np.array([0.0], dtype=float)
        C_eff = (float(self.C) / n) if self.scale_C else float(self.C)
        lb = np.zeros(n, dtype=float)
        ub = np.full(n, C_eff, dtype=float)

        # Solve (reuse your quadprog function from SVM)
        alpha, _ = quadprog_gurobi(H, f, Aeq, beq, lb, ub, verbose=self.verbose)
        self.alpha_ = alpha

        # Identify support “vectors” (bags)
        sv_mask = alpha > self.tol
        self.support_ = np.flatnonzero(sv_mask).astype(int)
        self.support_vectors_ = [bag_list[i] for i in self.support_]   # store the Bag objects
        self.dual_coef_ = (alpha[sv_mask] * Y[sv_mask]).reshape(1, -1)

        # Intercept from margin SVs (0 < α_i < C_eff)
        on_margin = (alpha > self.tol) & (alpha < C_eff - self.tol)
        if not np.any(on_margin):
            on_margin = sv_mask
        b_vals = Y[on_margin] - (alpha * Y) @ K[:, on_margin]
        self.intercept_ = float(np.mean(b_vals)) if b_vals.size else 0.0

        # Linearization (recover w and recompute b from primal if possible)
        if self._can_linearize(bk):
            Z = np.stack(
                [self._phi(b, normalizer=bk.normalizer) for b in bag_list],
                axis=0
            )  # (n_bags, d)

            w = (self.alpha_ * self.y_) @ Z
            self._train_embeddings_ = Z
            self.coef_ = w

            # Recompute b using margin SVs (or all SVs if none on-margin)
            on_margin = (self.alpha_ > self.tol) & (self.alpha_ < C_eff - self.tol)
            use = on_margin if np.any(on_margin) else (self.alpha_ > self.tol)
            if np.any(use):
                b_vals = self.y_[use] - Z[use] @ w
                self.intercept_ = float(np.mean(b_vals))
        else:
            self.coef_ = None

        self.X_ = None
        return self
    
    
    def _phi(self, bag: Bag, *, normalizer: str) -> np.ndarray:
        """
        Bag embedding φ(B) in ℝ^d matching the bag kernel:
          - weights: uniform if use_intra=False, else normalized mask
          - normalizer:
              "none"         -> φ = mean
              "average"      -> φ = mean / count   (count = n or sum(mask))
              "featurespace" -> φ = mean / ||mean|| (for Linear instance kernel)
        """
        n, X = bag.n, bag.X
        if n == 0:
            return np.zeros((self.support_vectors_[0].d,), dtype=float)  # d from any train bag

        w = np.full(n, 1.0 / n, dtype=float)

        mean = (w[None, :] @ X).ravel()        # (d,)

        if normalizer == "none":
            return mean
        elif normalizer == "average":
            count = float(bag.n)
            return mean / max(count, 1e-12)
        elif normalizer == "featurespace":
            # For Linear instance kernel, FS-norm == ||mean||
            denom = float(np.linalg.norm(mean))
            return mean / max(denom, 1e-12)
        else:
            return mean  # fallback (shouldn't happen)

    def decision_function(self, bags):
        if self.bags_ is None or self.alpha_ is None or self.y_ is None or self.intercept_ is None:
            raise RuntimeError("Model is not fitted yet.")

        # Coerce to list of Bag
        if isinstance(bags, BagDataset):
            test_bags = list(bags.bags)
        elif len(bags) > 0 and isinstance(bags[0], Bag):  # type: ignore[index]
            test_bags = list(bags)  # type: ignore[assignment]
        else:
            test_bags = [Bag(X=np.asarray(b, dtype=float), y=0.0) for b in bags]

        bk = self._ensure_bag_kernel()

        # ---- if the coef_ exists (then fallback here)
        if self.coef_ is not None and self._can_linearize(bk):
            Zt = np.stack(
                [self._phi(b, normalizer=bk.normalizer) for b in test_bags],
                axis=0
            )
            return (Zt @ self.coef_ + self.intercept_).ravel()

        # fallback: kernel path
        Ktest = bk(self.bags_, test_bags)  # (n_train, n_test)
        return ((self.alpha_ * self.y_) @ Ktest + self.intercept_).ravel()

    # Optional: make predict/score robust to passing BagDataset or raw arrays
    def predict(self, bags: Sequence[Bag] | BagDataset | Sequence[np.ndarray]) -> npt.NDArray[np.float64]:
        scores = self.decision_function(bags)
        # Map sign back to original class order: classes_[0] ↔ -1, classes_[1] ↔ +1
        return np.where(scores >= 0.0, self.classes_[1], self.classes_[0])

    def score(self, bags, y_true) -> float:
        y_pred = self.predict(bags)
        y_true = np.asarray(y_true, dtype=float).ravel()
        return float((y_pred == y_true).mean())




__all__ = [
    "SVM", "NSK", "sMIL", "sAwMIL",
    "BaseKernel", "Linear", "RBF", "Polynomial", "Sigmoid",
    "Normalize", "Scale", "Sum", "Product",
    "WeightedMeanBagKernel", "make_bag_kernel",
    "Bag", "BagDataset",
]

__version__ = "0.1.3"

from .svm import SVM
from .nsk import NSK
from .smil import sMIL
from .sawmil import sAwMIL

from .kernels import (
    BaseKernel, Linear, RBF, Polynomial, Sigmoid,
    Normalize, Scale, Sum, Product,
)

from .bag_kernels import WeightedMeanBagKernel, make_bag_kernel
from .bag import Bag, BagDataset


from __future__ import annotations
from typing import Optional
from sklearn.base import BaseEstimator
import numpy as np
import numpy.typing as npt

from .kernels import get_kernel, BaseKernel, KernelType, Linear
from .quadprog import quadprog_gurobi

class SVM(BaseEstimator):
    """Support Vector Machine (dual QP w/ Gurobi). Binary only."""

    def __init__(
        self,
        C: float = 1.0,
        kernel: KernelType = "linear",
        gamma: Optional[float] = None,
        degree: int = 3,
        coef0: float = 0.0,
        tol: float = 1e-6,       # threshold to detect support vectors
        verbose: bool = False,
    ):
        self.C = C
        self.kernel = kernel
        self.gamma = gamma
        self.degree = degree
        self.coef0 = coef0
        self.tol = tol
        self.verbose = verbose

        # Fitted attributes
        self._k: Optional[BaseKernel] = None # the fitted kernel instance
        self.X_: Optional[npt.NDArray[np.float64]] = None
        self.y_: Optional[npt.NDArray[np.float64]] = None   # mapped to {-1,+1}
        self.classes_: Optional[npt.NDArray[np.float64]] = None  # original labels
        self.alpha_: Optional[npt.NDArray[np.float64]] = None
        self.support_: Optional[npt.NDArray[np.int64]] = None
        self.support_vectors_: Optional[npt.NDArray[np.float64]] = None
        self.dual_coef_: Optional[npt.NDArray[np.float64]] = None  # shape (1, n_SV)
        self.intercept_: Optional[float] = None
        self.coef_: Optional[npt.NDArray[np.float64]] = None  # only for linear kernel

    def _get_kernel(self, X_train: npt.NDArray[np.float64]) -> BaseKernel:
        """Instantiate and fit kernel on training X for defaults like gamma."""
        if self._k is None:
            self._k = get_kernel(self.kernel, gamma=self.gamma, degree=self.degree, coef0=self.coef0)
            self._k.fit(X_train)
        return self._k

    def fit(self, X: npt.NDArray[np.float64], y: npt.NDArray[np.float64]) -> "SVM":
        X = np.asarray(X, dtype=float)
        y = np.asarray(y, dtype=float).ravel()
        if X.ndim != 2:
            raise ValueError("X must be 2D.")
        if y.ndim != 1 or y.shape[0] != X.shape[0]:
            raise ValueError("y must be 1D with len(y) == n_samples.")

        # Map arbitrary binary labels to {-1,+1} (stable, order by np.unique)
        classes = np.unique(y)
        if classes.size != 2:
            raise ValueError("Binary classification only (exactly two classes required).")
        self.classes_ = classes.astype(float)
        y_mapped = np.where(y == classes[0], -1.0, 1.0)

        self.X_ = X
        self.y_ = y_mapped

        # Build dual QP pieces
        k = self._get_kernel(X)
        K = k(X, X)  # (n,n)
        Y = y_mapped
        H = (Y[:, None] * Y[None, :]) * K
        H = 0.5 * (H + H.T)  # symmetrize

        n = X.shape[0]
        f = -np.ones(n, dtype=float)
        Aeq = Y.reshape(1, -1)
        beq = np.array([0.0], dtype=float)
        lb = np.zeros(n, dtype=float)
        ub = np.full(n, float(self.C), dtype=float)

        # Solve dual
        alpha, _ = quadprog_gurobi(H, f, Aeq, beq, lb, ub, verbose=self.verbose)
        self.alpha_ = alpha

        # Support vectors
        sv_mask = alpha > self.tol
        self.support_ = np.flatnonzero(sv_mask).astype(int)
        self.support_vectors_ = X[sv_mask]
        self.dual_coef_ = (alpha[sv_mask] * Y[sv_mask]).reshape(1, -1)

        # Intercept b using margin SVs: 0 < α_i < C
        on_margin = (alpha > self.tol) & (alpha < self.C - self.tol)
        if not np.any(on_margin):  # degenerate case: use all SVs
            on_margin = sv_mask
        b_vals = Y[on_margin] - (alpha * Y) @ K[:, on_margin]
        self.intercept_ = float(np.mean(b_vals)) if b_vals.size else 0.0

        # Linear primal weights if kernel is strictly linear
        self.coef_ = None
        if isinstance(k, Linear):
            self.coef_ = (alpha * Y) @ X  # shape (n_features,)

        return self

    def decision_function(self, X: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        if self.X_ is None or self.alpha_ is None or self.y_ is None or self.intercept_ is None:
            raise RuntimeError("Model is not fitted yet.")
        X = np.asarray(X, dtype=float)
        # Fast path for linear kernel
        if self.coef_ is not None:
            return (X @ self.coef_) + self.intercept_
        k = self._get_kernel(self.X_)   # do NOT refit on X; use the training-fitted kernel
        Ktest = k(self.X_, X)           # (n_train, n_test)
        return (self.alpha_ * self.y_) @ Ktest + self.intercept_

    def predict(self, X: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        scores = self.decision_function(X)
        pred_sign = np.where(scores >= 0, 1.0, -1.0)
        if self.classes_ is None:
            return pred_sign
        y0, y1 = self.classes_
        return np.where(pred_sign < 0, y0, y1)

    def score(self, X: npt.NDArray[np.float64], y: npt.NDArray[np.float64]) -> float:
        y = np.asarray(y).ravel()
        yhat = self.predict(X)
        return float(np.mean(yhat == y))
